from setuptools import setup

setup(
   name='uaibot',
   version='0.1',
   description='A useful module',
   author='Vinicius Mariano and Johnata Brayan',
   author_email='foomail@foo.com',
   packages=['foo'],  #same as name
   install_requires=['quadprog', 'colour']
)